/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef POINTIMPL_HH
#define POINTIMPL_HH

#include <cmath>

namespace SharpConstruct
{
	Point3D::Point3D()
	: x_( 0 ), y_( 0 ), z_( 0 )
	{}
	Point3D::Point3D( const float x, const float y, const float z )
	: x_( x ), y_( y ), z_( z )
	{}

	float& Point3D::X()
	{
		return x_;
	}
	float& Point3D::Y()
	{
		return y_;
	}
	float& Point3D::Z()
	{
		return z_;
	}
	
	float Point3D::X() const
	{
		return x_;
	}
	float Point3D::Y() const
	{
		return y_;
	}
	float Point3D::Z() const
	{
		return z_;
	}
	
	void Point3D::operator+=( const Point3D& a )
	{
		x_ += a.x_;
		y_ += a.y_;
		z_ += a.z_;
	}
	void Point3D::operator-=( const Point3D& a )
	{
		x_ -= a.x_;
		y_ -= a.y_;
		z_ -= a.z_;
	}
	void Point3D::operator*=( const Point3D& a )
	{
		x_ *= a.x_;
		y_ *= a.y_;
		z_ *= a.z_;
	}
	void Point3D::operator/=( const Point3D& a )
	{
		x_ /= a.x_;
		y_ /= a.y_;
		z_ /= a.z_;
	}
	void Point3D::operator+=( const float a )
	{
		x_ += a;
		y_ += a;
		z_ += a;
	}
	void Point3D::operator-=( const float a )
	{
		x_ -= a;
		y_ -= a;
		z_ -= a;
	}
	void Point3D::operator*=( const float a )
	{
		x_ *= a;
		y_ *= a;
		z_ *= a;
	}
	void Point3D::operator/=( const float a )
	{
		x_ /= a;
		y_ /= a;
		z_ /= a;
	}
	
	Point3D Point3D::operator+( const Point3D& a ) const
	{
		return Point3D( x_ + a.x_, y_ + a.y_, z_ + a.z_ );
	}
	Point3D Point3D::operator-( const Point3D& a ) const
	{
		return Point3D( x_ - a.x_, y_ - a.y_, z_ - a.z_ );
	}
	Point3D Point3D::operator*( const Point3D& a ) const
	{
		return Point3D( x_ * a.x_, y_ * a.y_, z_ * a.z_ );
	}
	Point3D Point3D::operator/( const Point3D& a ) const
	{
		return Point3D( x_ / a.x_, y_ / a.y_, z_ / a.z_ );
	}
	Point3D Point3D::operator+( const float a ) const
	{
		return Point3D( x_ + a, y_ + a, z_ + a );
	}
	Point3D Point3D::operator-( const float a ) const
	{
		return Point3D( x_ - a, y_ - a, z_ - a );
	}
	Point3D Point3D::operator*( const float a ) const
	{
		return Point3D( x_ * a, y_ * a, z_ * a );
	}
	Point3D Point3D::operator/( const float a ) const
	{
		return Point3D( x_ / a, y_ / a, z_ / a );
	}

	Point3D Cross( const Point3D& a, const Point3D& b )
	{
		return Point3D( a.Y() * b.Z() - a.Z() * b.Y(),
				a.Z() * b.X() - a.X() * b.Z(),
				a.X() * b.Y() - a.Y() * b.X() );
	}

	float Distance( const Point3D& a, const Point3D& b )
	{
		const Point3D dif( b - a );

		return sqrt( HorizontalAdd( dif * dif ) );
	}

	float Dot( const Point3D& a, const Point3D& b )
	{
		return a.X() * b.X() + a.Y() * b.Y() + a.Z() * b.Z();
	}

	float HorizontalAdd( const Point3D& a )
	{
		return a.X() + a.Y() + a.Z();
	}

	Point3D Midpoint( const Point3D& a, const Point3D& b )
	{
		return Point3D( ( a + b ) / 2 );
	}

	void Normalize( Point3D& a )
	{
		a /= sqrt( HorizontalAdd( a * a ) );
	}

	Point3D Origin()
	{
		return Point3D( 0, 0, 0 );
	}

	Point3D PlaneNormal( const Point3D& a, const Point3D& b, const Point3D& c )
	{
		Point3D norm( Cross( a - b, c - b ) );
		Normalize( norm );
		return norm;
	}
	Point3D PlaneUnormal( const Point3D& a, const Point3D& b, const Point3D& c )
	{
		return Cross( a - b, c - b );
	}
}

#endif
