/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* This file uses code from the tutorial at
   http://www.cse.unsw.edu.au/~lambert/splines/
*/

#include "Curve.hh"

#include <iostream>

#ifdef WIN32
#include "win32.h"
#endif

using SharpConstruct::Curve;
using SharpConstruct::CurveEditor;

Curve::Curve()
{
	AddNode( 0, 0 );
	AddNode( 0, 0 );
	AddNode( 1.0f / 3.0f, 0.15 );
	AddNode( 2.0f / 3.0f, 0.75 );
	AddNode( 1, 1 );
	AddNode( 1, 1 );
}

const Curve::List& Curve::Nodes() const
{
	return nodes_;
}
void Curve::SetNode( const unsigned i, const Node& loc )
{
	Node cp( loc );

	if( cp.X() < 0 )
		cp.SetX( 0 );
	if( cp.Y() < 0 )
		cp.SetY( 0 );
	if( cp.X() > 1 )
		cp.SetX( 1 );
	if( cp.Y() > 1 )
		cp.SetY( 1 );

	// TC: Added cast because ( 0UL -1 ) == 0xffffffffUL, which is always >= 0.
	// Doing this correctly requires signed arithmetics.
	if( ( static_cast<signed int>( i - 1 ) >= 0 ) && ( cp.X() < nodes_[ i - 1 ].X() ) )
		cp.SetX( nodes_[ i - 1 ].X() );
	if( ( i + 1 < nodes_.size() ) && ( cp.X() > nodes_[ i + 1 ].X() ) )
		cp.SetX( nodes_[ i + 1 ].X() );

	if( i < 2 )
		cp.SetX( 0 );
	if( i > nodes_.size() - 3 )
		cp.SetX( 1 );

	if( i < 2 )
		nodes_[ 0 ] = nodes_[ 1 ] = cp;
	else if( i > nodes_.size() - 3 )
		nodes_[ nodes_.size() - 2 ] =
			nodes_[ nodes_.size() - 1 ] = cp;
	else
		nodes_[ i ] = cp;
}

int Curve::AddNode( const Node& n )
{
	return AddNode( n.X(), n.Y() );
}

int Curve::AddNode( float x, float y )
{
	/*if( x < 0 )
		x = 0;
	if( x > 1 )
		x = 1;
	if( y < 0 )
		y = 0;
	if( y > 1 )
		y = 1;

	nodes_.push_front( Node( x, y ) );
	nodes_.sort( NodeSort() );*/

	const Node node( x, y );

	nodes_.push_back( node );
	sort( nodes_.begin(), nodes_.end(), NodeSort() );

	Update();

	for( unsigned i = 0; i < nodes_.size(); ++i )
		if( nodes_[ i ] == node )
			return i;
	return -1;
}

void Curve::RemoveNode( const unsigned i )
{
	nodes_.erase( find( nodes_.begin(), nodes_.end(), nodes_[ i ] ) );
}

const Curve::List& Curve::RenderCache() const
{
	return render_cache_;
}

void Curve::Update()
{
	const unsigned steps( 20 );

	List& line( render_cache_ );
	line.clear();

	line.push_back( interpolate_( 2, 0 ) );
	float prev( 0 );
	for( unsigned i = 2; i < nodes_.size() - 1; ++i )
	{
		for( unsigned j = 1; j <= steps; j++ )
		{
			Curve::Node cur( interpolate_( i, j * 1.0f / steps ) );
			if( cur.X() < prev )
				cur.SetX( prev );
			line.push_back( cur );
			prev = cur.X();
		}
	}
}

float Curve::Value( const float x ) const
{
	for( unsigned i = 0; i < render_cache_.size(); ++i )
		if( render_cache_[ i ].X() >= x )
			return render_cache_[ i ].Y();
	return 1;
}

bool Curve::NodeSort::operator()( const Node& a,
				  const Node& b ) const
{
	if( a.X() == b.X() )
		return a.Y() < b.Y();
	else
		return a.X() < b.X();
}

Curve::Node Curve::interpolate_( const int i, const float t ) const
{
	float px( 0 );
	float py( 0 );
	for( int j = -2; j <= 1; ++j )
	{
		if( ( i + j ) < static_cast< signed >( nodes_.size() ) )
		{
			px += evaluate_( j, t ) * nodes_[ i + j ].X();
			py += evaluate_( j, t ) * nodes_[ i + j ].Y();
		}
	}

	return Node( px, py );
}

float Curve::evaluate_( const int i, const float t ) const
{
	if( i == -2 )
		return ((-t+2)*t-1)*t/2;
	else if( i == -1 )
		return (((3*t-5)*t)*t+2)/2;
	else if( i == 0 )
		return ((-3*t+4)*t+1)*t/2;
	else if( i == 1 )
		return ((t-1)*t*t)/2;
	else
		return 0;
}

CurveEditor::CurveEditor()
: selected_node_( -1 )
{
	add_events( Gdk::EXPOSURE_MASK | Gdk::POINTER_MOTION_MASK |
		    Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK );

	set_size_request( -1, 100 );
	show();
}

const Curve& CurveEditor::Data() const
{
	return curve_;
}

bool CurveEditor::on_expose_event( GdkEventExpose* e )
{
	using namespace Gdk;

	Glib::RefPtr< GC > gc( GC::create( get_window() ) );

	const Curve::List render( curve_.RenderCache() );
	const int height( get_allocation().get_height() );
	for( Curve::List::const_iterator i = render.begin();
	     i != render.end(); ++i )
	{
		Curve::List::const_iterator next( i );
		++next;
		if( next != render.end() )
		{
			const Gdk::Point p1( project_( *i ) );
			const Gdk::Point p2( project_( *next ) );

			std::vector< Point > poly_points;
			poly_points.push_back( p1 );
			poly_points.push_back( p2 );
			poly_points.push_back( Gdk::Point( p2.get_x(), height ) );
			poly_points.push_back( Gdk::Point( p1.get_x(), height ) );

			gc->set_rgb_fg_color( Color( "#888888" ) );
			get_window()->draw_polygon( gc, true, poly_points );

			gc->set_rgb_fg_color( Color( "Black" ) );
			get_window()->draw_line( gc, p1.get_x(), p1.get_y(), p2.get_x(), p2.get_y() );
		}
	}

	gc->set_rgb_fg_color( Color( "Black" ) );
	const Curve::List& cl( curve_.Nodes() );
	for( Curve::List::const_iterator i = cl.begin(); i != cl.end(); ++i )
	{
		const Gdk::Point loc( project_( *i ) );
		get_window()->draw_arc( gc, true, loc.get_x() - RADIUS,
					loc.get_y() - RADIUS, DIAMETER,
					DIAMETER, 0, 360 * 64 );
	}
	
	return true;
}

bool CurveEditor::on_button_press_event( GdkEventButton* e )
{
	const Gdk::Point loc( lroundf( e->x ), lroundf( e->y ) );
	const Curve::List& cl( curve_.Nodes() );
	for( unsigned i = 0; i < cl.size(); ++i )
	{
		const Gdk::Point node_loc( project_( cl[ i ] ) );
		if( Math::Distance( loc, node_loc ) < 8 )
		{
			selected_node_ = i;
			break;
		}
	}

	if( e->button == 3 || e->state & GDK_SHIFT_MASK )
	{
		if( selected_node_ != -1 )
		{
			curve_.RemoveNode( selected_node_ );
			selected_node_ = -1;
			curve_.Update();
			queue_draw();
		}
	}
	else if( selected_node_ == -1 )
	{
		selected_node_ = curve_.AddNode( unproject_( loc ) );
		curve_.Update();
		queue_draw();
	}

	return true;
}
bool CurveEditor::on_motion_notify_event( GdkEventMotion* e )
{
	if( selected_node_ != -1 )
	{
		const Curve::Node loc( unproject_( Gdk::Point( lroundf( e->x ) ,lroundf( e->y ) ) ) );
		curve_.SetNode( selected_node_, loc );
		curve_.Update();
	}

	queue_draw();
	
	return true;
}
bool CurveEditor::on_button_release_event( GdkEventButton* )
{
	selected_node_ = -1;

	return true;
}

Gdk::Point CurveEditor::project_( const Curve::Node& n ) const
{
	return Gdk::Point( lroundf( n.X() * ( get_allocation().get_width() - 4 ) + 2 ),
			   lroundf( ( 1 - n.Y() ) * ( get_allocation().get_height() - 4 ) + 2 ) );
}

Curve::Node CurveEditor::unproject_( const Gdk::Point& p ) const
{
	return Curve::Node( ( ( float )p.get_x() - 2 ) / ( get_allocation().get_width() - 4 ),
			    1 - ( ( ( float )p.get_y() - 2 ) / ( get_allocation().get_height() - 4 ) ) );
}
