/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "DisplaceBrush.hh"
#include "Mesh.hh"

#include <iostream>

using SharpConstruct::Brush::Displace;
using SharpConstruct::Brush::EditData;

SharpConstruct::Brush::SupportedProperties Displace::Properties() const
{
	return SupportedProperties( true, true, true, true, true, true, true,
				    true, true );
}

void Displace::pre_start_sculpt_()
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	mesh.ModifiedVertexIndices().clear();

	if( Layer() )
	{
		displacements_.clear();
		color_adds_.clear();
		if( Direction() != OFF )
			displacements_.resize( mesh.VisibleElements().Vertices, 0 );
		if( Colorize() )
			color_adds_.resize( mesh.VisibleElements().Vertices, 0 );
	}
}

void Displace::simple_displace_( const EditData& e )
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	mesh.FindActiveVertices( active_data_, e );

	if( Mask() )
		mask_();
	else
	{
		if( Direction() != OFF )
		{
			const Point3D area_normal( mesh.AreaNormal( e, active_data_ ) );
			for( unsigned i = 0; i < active_data_.size(); ++i )
			{
				const ActiveData& a( active_data_[ i ] );
				Point3D& loc( mesh.VertexLocations()[ a.Index() ] );

				loc += area_normal * a.Strength() * Strength() * mesh.Mask( a.Index() );
			}
		}
		if( Colorize() )
		{
			for( unsigned i = 0; i < active_data_.size(); ++i )
			{
				const ActiveData& a( active_data_[ i ] );
				SharpConstruct::Color& v( mesh.VertexColors()[ a.Index() ] );
	
				const float intensity( a.Strength() * Intensity() );
				v += ( Pigment() - v ) * intensity * mesh.Mask( a.Index() );	
			}
		}
	}
}

void Displace::layer_displace_( const EditData& e )
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	mesh.FindActiveVertices( active_data_, e );

	if( Colorize() )
	{
		std::vector< Color >& colbackup = MeshHistory::Instance().GetPreviousComposite().Colors()->VertexColors();
		for( unsigned i = 0; i < active_data_.size(); ++i )
		{
			const ActiveData& a( active_data_[ i ] );

			const float colval = a.Strength() * Intensity();
			{
				color_adds_[ a.Index() ] += colval;
				if( color_adds_[ a.Index() ] > Intensity() )
					color_adds_[ a.Index() ] = Intensity();

				// Revert vertex color
				mesh.VertexColors()[ a.Index() ] = colbackup[ a.Index() ];

				mesh.VertexColors()[ a.Index() ] +=
					( Pigment() - colbackup[ a.Index() ] ) *
					color_adds_[ a.Index() ] * mesh.Mask( a.Index() );
			}
		}
	}

	if( Direction() != OFF )
	{
		Point3DVector& locbackup = MeshHistory::Instance().GetPreviousComposite().Vertices()->VertexLocations();
		const Normal3D area_normal( mesh.AreaNormal( e, active_data_ ) );
		for( unsigned i = 0; i < active_data_.size(); i++ )
		{
			const ActiveData& a( active_data_[ i ] );

			const float addval( active_data_[ i ].Strength() * Strength() * mesh.Mask( a.Index() ) );

			{
				float& disp( displacements_[ a.Index() ] );

				// First 'if' insures that we don't decrease the displacement at end of stroke
				if( ( Strength() > 0 && disp < Strength() ) ||
				    ( Strength() < 0 && disp > Strength() ) )
				{
					disp += addval;
					
					if( Strength() < 0 )
					{
						if( disp < Strength() )
							disp = Strength();
					}
					else
					{
						if( disp > Strength() )
							disp = Strength();
					}
					
					mesh.VertexLocations()[ a.Index() ] = locbackup[ a.Index() ] +
						( area_normal * disp );
				}
			}
		}
	}
}

void Displace::mask_()
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	for( unsigned i = 0; i < active_data_.size(); ++i )
	{
		const ActiveData& a( active_data_[ i ] );
		const float intensity( a.Strength() * Intensity() );
		SharpConstruct::Color& v( mesh.VertexColors()[ a.Index() ] );

		float goal( 1 );
		if( Direction() == IN )
			goal = 0;
		v.SetMask( v.Mask() + ( goal - v.Mask() ) * intensity );
	}
}

float Displace::factor_() const
{
	return 10000;
}

void Displace::pre_continue_sculpt_()
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	if( Dot() )
	{
		const Point3DVector& locbackup( MeshHistory::Instance().GetPreviousComposite().Vertices()->VertexLocations() );
		const Point3DVector& norbackup( MeshHistory::Instance().GetPreviousComposite().Vertices()->VertexNormals() );
		const std::vector< Color >& colbackup( MeshHistory::Instance().GetPreviousComposite().Colors()->VertexColors() );
		const std::vector< unsigned >& modvi( mesh.ModifiedVertexIndices() );

		for( unsigned i = 0; i < modvi.size(); ++i )
		{
			if( Direction() != OFF )
			{
				mesh.VertexLocations()[ modvi[ i ] ] =
					locbackup[ modvi[ i ] ];
				mesh.VertexNormals()[ modvi[ i ] ] =
					norbackup[ modvi[ i ] ];
			}
			if( Colorize() )
				mesh.VertexColors()[ modvi[ i ] ] =
					colbackup[ modvi[ i ] ];
		}
	}
	else
		mesh.ModifiedVertexIndices().clear();
}

void Displace::start_sculpt_( const EditData& e )
{
	if( !Layer() )
		simple_displace_( e );
}

void Displace::continue_sculpt_( const EditData& e )
{
	if( Layer() )
		layer_displace_( e );
	else
		simple_displace_( e );
}
